from datetime import datetime
from bson.objectid import ObjectId


class BaseField:
    def __init__(self, value=None):
        self._value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value


class MongoIDField(BaseField):
    @property
    def value(self):
        return str(self._value) if self._value else None

    @value.setter
    def value(self, value):
        if value:
            self._value = ObjectId(value)
        else:
            raise Exception("MongoIDField expects a value")


class StringField(BaseField):
    @property
    def value(self):
        return str(self._value) if self._value else ""

    @value.setter
    def value(self, value):
        self._value = str(value) if value else ""


class IntegerField(BaseField):
    @property
    def value(self):
        return int(self._value) if self._value else 0

    @value.setter
    def value(self, value):
        if value is not None:
            if isinstance(value, int):
                self._value = value
            else:
                raise ValueError("IntegerField expects an integer")
        else:
            self._value = 0


class BooleanField(BaseField):
    @property
    def value(self):
        return bool(self._value)

    @value.setter
    def value(self, value):
        if value is not None:
            if isinstance(value, bool):
                self._value = value
            else:
                raise ValueError("BooleanField expects a boolean")
        else:
            self._value = None


class FloatField(BaseField):
    @property
    def value(self):
        return float(self._value) if self._value else 0.0

    @value.setter
    def value(self, value):
        if value is not None:
            if isinstance(value, float):
                self._value = value
            else:
                raise ValueError("FloatField expects a float")
        else:
            self._value = 0.0


class MapField(BaseField):
    @property
    def value(self):
        return dict(self._value) if self._value else {}

    @value.setter
    def value(self, value):
        if value is not None:
            if value and isinstance(value, dict):
                self._value = value
            else:
                raise ValueError("MapField expects a dict")
        else:
            self._value = {}


class ListField(BaseField):
    @property
    def value(self):
        v = super().value
        if v:
            return list(v)
        else:
            return []

    @value.setter
    def value(self, value):
        if value is not None:
            if value and isinstance(value, list):
                self._value = value
            else:
                raise ValueError("ListField expects a list")
        else:
            self._value = []


class DateTimeField(BaseField):
    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, value):
        if value is not None:
            if isinstance(value, datetime):
                self._value = value
            else:
                raise ValueError("DateTimeField expects a datetime object")
        else:
            self._value = None


class TimestampField(BaseField):
    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, value):
        if value is not None:
            if isinstance(value, int):
                self._value = value
            else:
                raise ValueError("DateTimeField expects a datetime object")
        else:
            self._value = None


class EmbeddedDocumentField(BaseField):
    def __init__(self, embedded_model, value=None):
        self.embedded_model = embedded_model
        super().__init__(value)

    @property
    def value(self):
        v = super().value
        if v:
            return self.embedded_model(v).to_dict()
        else:
            return None

    @value.setter
    def value(self, value):
        if value is not None:
            if isinstance(value, dict):
                self._value = self.embedded_model(value)
            else:
                raise ValueError("EmbeddedDocumentField expects a dict")
        else:
            self._value = None


class ReferenceField(BaseField):
    @property
    def value(self):
        if self._value:
            return str(self._value)
        return None

    @value.setter
    def value(self, value):
        if value is not None:
            if isinstance(value, str):
                self._value = ObjectId(value)
            else:
                raise ValueError("ReferenceField expects an ObjectId")
        else:
            self._value = None



from .utils.normalizer import MongoNormalizer
from .utils.func import cal_timestamp
from .field import *
from bson.objectid import ObjectId
from .index_store import IndexStore, create_index_hash


class ValueErrorException(Exception):
    pass


DATA_TYPE_MAPPINGS = {
    'MongoIDField': MongoIDField,
    'FloatField': FloatField,
    'IntegerField': IntegerField,
    'StringField': StringField,
    'BooleanField': BooleanField,
    'MapField': MapField,
    'DateTimeField': DateTimeField,
    'EmbeddedDocumentField': EmbeddedDocumentField,
    'ReferenceField': ReferenceField,
    'ListField': ListField,

}


class BaseModel(object):
    indexes = [
        [('changed', -1)],
    ]

    def __init__(self, collection, config=None):
        self.id = MongoIDField()
        self.deleted = BooleanField(False)
        self.created = TimestampField()
        self.changed = TimestampField()

        self._collection = collection
        self.config = config

        self.store = IndexStore()

        if self.indexes:
            try:
                self._prepare_indexes(self._collection)
            except Exception as e:
                # raise Exception(f"Error creating indexes: {e}")
                # TODO: log error or do something more useful
                pass

        if self.id:
            self.initialize()

    def add_indexes(self, indexes):
        if isinstance(indexes, list):
            unique_indexes = [index for index in indexes if str(index) not in map(str, self.indexes)]
            self.indexes.extend(unique_indexes)

    def _prepare_indexes(self, collection):
        computed_hash = create_index_hash(self.indexes)
        stored_hash = self.store.get_hash(collection.name)
        if computed_hash != stored_hash:
            print('dropping indexes')
            collection.drop_indexes()

        self.store.set_hash(collection.name, computed_hash)

        existing_indexes = collection.list_indexes()
        existing_index_names = [index['name'] for index in existing_indexes]

        for index_definition in self.indexes:
            index_keys = [(field, _) for field, _ in index_definition[:-1]] if isinstance(index_definition[-1],
                                                                                          dict) else [
                (field, _) for field, _ in index_definition]

            index_options = index_definition[-1] if isinstance(index_definition[-1],
                                                               dict) else {}
            index_name = "_".join([f"{field}_{_}" for field, _ in index_keys])

            if index_name not in existing_index_names:
                collection.create_index(index_keys, **index_options)

    def convert_value_by_field_type(self, field_name, value):
        field_class = type(self.__dict__[field_name])
        field_object = field_class(value)
        return field_object.value

    def __setattr__(self, key, value):
        try:
            attr = object.__getattribute__(self, key)
            if attr and isinstance(attr, BaseField):
                try:
                    update_func = '{}_alter_value'.format(key)
                    # if object.__getattribute__(self, update_func):
                    if hasattr(self, update_func):
                        value = getattr(self, update_func)(value)
                except Exception as e:
                    raise ValueErrorException(e)

                if isinstance(attr, MongoIDField):
                    object.__setattr__(self, key, MongoIDField(value))
                else:
                    field_type_class = type(attr).__name__
                    field_class = DATA_TYPE_MAPPINGS.get(field_type_class)
                    if field_class:
                        value = field_class(value)
                    else:
                        value = BaseField(value)

                    object.__setattr__(self, key, value)
        except Exception as e:
            object.__setattr__(self, key, value)

            # exception is of type ValueErrorException then raise it
            if isinstance(e, ValueErrorException):
                raise e

    def __getattribute__(self, key):
        try:
            attr = object.__getattribute__(self, key)
            if attr and isinstance(attr, BaseField):
                return attr.value

        except:
            pass

        return object.__getattribute__(self, key)

    def get(self, model_id):
        model = self._collection.find_one({'_id': ObjectId(model_id)})
        return self.initialize(model)

    def get_id(self):
        return ObjectId(self.id)

    def save(self, data=None):
        self.set_bulk_fields(data)

        fields = {}
        attrs = self.__dict__
        for k, v in attrs.items():
            if isinstance(v, MongoIDField):
                fields[k] = ObjectId(v.value) if v.value else None
            elif isinstance(v, BaseField):
                fields[k] = v.value

        if fields:
            try:
                self._perform_insert_or_update(fields)
                return self.initialize()
            except Exception as e:
                raise Exception('{} could not be saved. Error: {}'.format(self.__class__.__name__, e))

    def _perform_insert_or_update(self, fields):
        fields['changed'] = cal_timestamp()

        # delete id in any case
        if 'id' in fields:
            del fields['id']

        try:
            if not self.id:
                fields['created'] = fields['changed']
                result = self._collection.insert_one(fields)
                self.id = str(result.inserted_id)
            else:
                result = self._collection.replace_one({'_id': ObjectId(self.id)}, fields)
        except Exception as e:
            raise Exception(e)

    def delete(self, soft_delete=False):
        if self.id:
            if not soft_delete:
                self._collection.delete_one({"_id": ObjectId(self.id)})
            else:
                self.deleted = True
                self.save()

    def initialize(self, model=None):
        if isinstance(model, BaseModel):
            return model

        if not model:
            model = self._collection.find_one({'_id': ObjectId(self.id)})

        if not model:
            raise Exception('Cannot initialize ' + self.__module__)

        normalized = MongoNormalizer.deserialize(model)
        for k, v in self.get_fields().items():
            # Check to prevent any missing/new column from being initialized
            if k in normalized or (k == 'id' and '_' + k in normalized):
                self.__dict__[k].value = normalized['_' + k if k == 'id' else k]

        return self

    def get_fields(self):
        fields = {}
        attrs = self.__dict__
        for k, v in attrs.items():
            if isinstance(v, BaseField):
                fields[k] = v.value

        return fields

    def set_bulk_fields(self, data):
        if data:
            for k in data:
                self.__setattr__(k, data[k])

    def __str__(self):
        return str(self.get_fields())

    @staticmethod
    def mongo_id(mongo_id):
        return ObjectId(mongo_id)

    def get_type(self):
        return type(self).__name__.lower()



from bson.json_util import dumps
from json import loads as json_loads


class MongoNormalizer:

    @classmethod
    def _handle_oid(cls, record):
        if '$oid' in record:
            return record['$oid']
        return record

    @classmethod
    def deserialize(cls, data, use_new_format=False):
        if not data:
            return [] if isinstance(data, list) else None

        records = json_loads(dumps(data))

        is_single_record = not isinstance(records, list)
        if is_single_record:
            records = [records]

        for record in records:
            for key, value in list(record.items()):
                if isinstance(value, dict):
                    record[key] = cls._handle_oid(value)

            if use_new_format and '_id' in record:
                record['id'] = record.pop('_id')

        return records[0] if is_single_record else records



import hashlib
import json


def create_index_hash(indexes):
    index_str = json.dumps(indexes, sort_keys=True)
    return hashlib.sha256(index_str.encode()).hexdigest()


INDEX_STORE_COLLECTION = None


class IndexStore(object):
    def __init__(self, collection=None):
        global INDEX_STORE_COLLECTION

        if collection is None:
            if INDEX_STORE_COLLECTION is None:
                raise ValueError("Index store collection not set")
            self.collection = INDEX_STORE_COLLECTION
        else:
            self.collection = collection
            INDEX_STORE_COLLECTION = collection

    def get_hash(self, collection_name):
        col = self.collection.find_one({'col_name': collection_name})
        if col:
            return col['hash']

        return None

    def set_hash(self, collection_name, index_hash):
        return self.collection.update_one({'col_name': collection_name}, {'$set': {'hash': index_hash}}, upsert=True)

    def delete_hash(self, collection_name):
        return self.collection.delete_one({'col_name': collection_name})
